MODULE EMC;
(* =========================================================================  
   Astrobe External Memory Controller (EMC) Initialisation Functions for SDRAM

   Target: 
     LPC40xx systems with an ISSI IS42S32800D 8M x 32 Synchronous DRAM connected
     to the Extended Memory Controller (EMC)
     
   Tested with:
     EA LPC4088 QuickStart board 
     
   Refs: 
     NXP UM10562 LPC408x/407x User manual
     ISSI IS42S32800D DataSheet
     Oberon for Cortex-M4 Microcontrollers
   
  (c) 2015 CFB Software   
  http://www.astrobe.com  
 
  ========================================================================= *)

IMPORT MCU, SYSTEM, Timer;

PROCEDURE* ConfigureAddr(addr: INTEGER; pins: SET);
VAR
  i: INTEGER;
  s: SET;
BEGIN
  (* Configure each pin included in the set with function 001: EMC *)
  FOR i := 0 TO 31 DO
    IF i IN pins THEN 
      SYSTEM.GET(addr, s);
      SYSTEM.PUT(addr, s - {1,2} + {0})
    END;
    INC(addr, 4); 
  END;
END ConfigureAddr;


PROCEDURE ConfigurePins;
BEGIN 
  (* CAS:P2.16, RAS:P2.17, CLKOUT0:P2.18, DYCS0:P2.20, CKE0:P2.24 *)
  (* DQM0 to DQM3 : P2.28 to P2.31 *)
  ConfigureAddr(MCU.IOCONP20, {16, 17, 18, 20, 24, 28..31});
  (* Data pins D0 to D31 : P3.0 to P3.31 *)
  ConfigureAddr(MCU.IOCONP30, {0..31});
  (* Address pins A0 to A14 : P4.0 to P4.14, WE: P4.25 *)
  ConfigureAddr(MCU.IOCONP40, {0..14, 25});
END ConfigurePins;
 

PROCEDURE Init*();
CONST
  SDRAMAddr = 0A0000000H;
  
  (* Power Control *)
  PCEMC = {11};
  
  (* Dynamic Configuration *)
  (* Address Mapping: 32-bits, 12 rows, 4 banks, 9 columns *)
  (* Bits 14 = 1 12 = 0 11:9 = 010 8:7 = 01 *)
  AM0 = {10, 7};
  AM1 = {14};
  B   = {19};
  
  (* Dynamic Control *)
  CE     = {0};
  CS     = {1};
  NORMAL = {};
  MODE   = {7};
  PALL   = {8};
  NOP    = {7,8};

VAR
  i, offset, modeReg, modeAddr, dummy: INTEGER;
  s: SET;
BEGIN
  SYSTEM.GET(MCU.PCONP, s);
  SYSTEM.PUT(MCU.PCONP, s + PCEMC);
  (* Clock is half CPU-rate *)
  SYSTEM.PUT(MCU.EMCCLKSEL, 1);
  (* Delays FBCCLKDLY = 16, CMDDLY = 16 *)
  SYSTEM.PUT(MCU.EMCDLYCTL, LSL(16, 8) + 16);
  (* Enabled *)
  SYSTEM.PUT(MCU.EMCCONTROL, 1);
  (* Little-endian *)
  SYSTEM.PUT(MCU.EMCCONFIG, 0);
  
  ConfigurePins();
  
  SYSTEM.PUT(MCU.EMCDYNCONFIG0, AM0 + AM1); 
 
  (* 60 / 72 / 80 MHz Bus *)
  (* RAS / CAS latency = 2 EMC CCLK cycles *)
  SYSTEM.PUT(MCU.EMCDYNRASCAS0, LSL(2, 8) + 2);
  SYSTEM.PUT(MCU.EMCDYNREADCONFIG, 1);
  SYSTEM.PUT(MCU.EMCDYNRP,   1);
  SYSTEM.PUT(MCU.EMCDYNRAS,  3);
  SYSTEM.PUT(MCU.EMCDYNSREX, 5);
  SYSTEM.PUT(MCU.EMCDYNAPR,  2);
  SYSTEM.PUT(MCU.EMCDYNDAL,  3);
  SYSTEM.PUT(MCU.EMCDYNWR,   1);
  SYSTEM.PUT(MCU.EMCDYNRC,   4);
  SYSTEM.PUT(MCU.EMCDYNRFC,  4);
  SYSTEM.PUT(MCU.EMCDYNXSR,  5);
  SYSTEM.PUT(MCU.EMCDYNRRD,  1);
  SYSTEM.PUT(MCU.EMCDYNMRD,  1);
 
  SYSTEM.PUT(MCU.EMCDYNCONTROL, CE + CS + NOP); 
  Timer.MSecDelay(200);     
  SYSTEM.PUT(MCU.EMCDYNCONTROL, CE + CS + PALL); 
  SYSTEM.PUT(MCU.EMCDYNREFRESH, 2);
  FOR i := 1 TO 080H DO END; (* 128 AHB clock cycles *)
 
  (* 60MHz Bus *)
  SYSTEM.PUT(MCU.EMCDYNREFRESH, 3AH); 
  
  (* Setup the mode register: UM10562 Section 9.11.1 *)
  SYSTEM.PUT(MCU.EMCDYNCONTROL, CE + CS + MODE);
  (* Burst length = 4: M1 = 1, CAS latency = 2: M5 = 1 *)
  modeReg := ORD({5, 1});
  (* Columns = 9, Bank bits = 2, Bus width / 16 = 2 *)
  offset := 9 + 2 + 2;
  modeAddr := SDRAMAddr + LSL(modeReg, offset);
  SYSTEM.GET(modeAddr, dummy);
  
  SYSTEM.PUT(MCU.EMCDYNCONTROL, NORMAL); 
  SYSTEM.PUT(MCU.EMCDYNCONFIG0, B + AM0 + AM1); 
 
END Init;
 
END EMC.
 
