MODULE SDRAMHeap;
(* =========================================================================  
   Astrobe External Memory Controller (EMC) Example Functions for SDRAM
   
   Description:
     1. Use the Oberon NEW function to allocate 1024 records each containing 
        an array of INTEGER with ~32Kb of data.
        Populate every element of every array with a value ranging from 0 to 
        1024 * 8191 - 1.
        Display how much heap memory is available after the allocation.
     2. Verify that none of the data has changed since it was written 
     3. Deallocate all the records using the Oberon DISPOSE function and display 
        how much heap memory is now available to be reused again.

   Target: 
     LPC40xx systems with an ISSI IS42S32800D 8M x 32 Synchronous DRAM connected
     to the Extended Memory Controller (EMC)
     
   Tested with:
     EA LPC4088 QuickStart board 
     
   Pre-requisites:
     The following values must be set in the Configuration file used to link the program:

       Heap Start: A0000000H
       Heap Limit: A2000000H
   
   Refs: 
     NXP UM10562 LPC408x/407x User manual
     ISSI IS42S32800D DataSheet
     Oberon for Cortex-M4 Microcontrollers
   
  (c) 2015 CFB Software   
  http://www.astrobe.com  
 
  ========================================================================= *)

IMPORT EMC, Main, Out, Storage, SYSTEM;

TYPE
  Data = ARRAY 8191 OF INTEGER;
  Buffer = 
    POINTER TO RECORD 
      data: Data
    END;

VAR
  buffers: ARRAY 1024 OF Buffer; 

   
PROCEDURE* InitData(VAR data: ARRAY OF INTEGER; VAR value: INTEGER);
VAR
  i: INTEGER;
BEGIN
  FOR i := 0 TO LEN(data) - 1 DO 
    data[i] := value;
    INC(value)
  END
END InitData;


PROCEDURE* VerifyData(VAR data: ARRAY OF INTEGER; VAR value: INTEGER);
VAR
  i: INTEGER;
BEGIN
  FOR i := 0 TO LEN(data) - 1 DO
    ASSERT(data[i] = value, 20); 
    INC(value)
  END 
END VerifyData;


PROCEDURE DisplayHeader(title: ARRAY OF CHAR);
BEGIN
  Out.String(title); Out.Ln;
  Out.String(" Index");
  Out.String("     Address");
  Out.String(" Available");
  Out.Ln
END DisplayHeader;


PROCEDURE DisplayStatus(index: INTEGER); 
BEGIN
  IF (index = 0) OR ((index+1) MOD 128 = 0) THEN 
    Out.Int(index, 6);
    Out.Hex(SYSTEM.ADR(buffers[index].data), 12); 
    Out.Int(Storage.HeapAvailable(), 10); 
    Out.Ln()
  END 
END DisplayStatus;

      
PROCEDURE Allocate;
VAR
  i, value: INTEGER;
BEGIN
  value := 0;
  FOR i := 0 TO LEN(buffers) - 1 DO
    NEW(buffers[i]);
    InitData(buffers[i].data, value);
    DisplayStatus(i)
  END
END Allocate;


PROCEDURE Verify;
VAR
  i, value: INTEGER;
BEGIN
  value := 0;
  FOR i := 0 TO LEN(buffers) - 1 DO
    VerifyData(buffers[i].data, value);
    DisplayStatus(i)
  END
END Verify;


PROCEDURE Deallocate;
VAR
  i: INTEGER;
BEGIN
  FOR i := LEN(buffers) - 1 TO 0 BY -1  DO
    DisplayStatus(i);
    DISPOSE(buffers[i]) 
  END;
END Deallocate;


BEGIN
  Out.String("SDRAM / Heap"); Out.Ln();
  Out.String("Initialising EMC..."); Out.Ln();
  EMC.Init();
  DisplayHeader("Allocate..."); 
  Allocate();
  DisplayHeader("Verify..."); 
  Verify();
  DisplayHeader("Deallocate..."); 
  Deallocate();
  Out.String("Finished"); Out.Ln();
END SDRAMHeap.
